@@ -25,6 +25,7 @@ import misc
 from ui import gnomeglade
 import melddoc
 import paths
+import ui.emblemcellrenderer
 import vc
 
 ################################################################################
@@ -33,16 +34,6 @@ import vc
 #
 ################################################################################
 
-def _expand_to_root( treeview, path ):
-    """Expand rows from path up to root"""
-    start = path[:]
-    while len(start) and not treeview.row_expanded(start):
-        start = start[:-1]
-    level = len(start)
-    while level < len(path):
-        level += 1
-        treeview.expand_row( path[:level], 0)
-
 def _commonprefix(files):
     if len(files) != 1:
         workdir = misc.commonprefix(files)
@@ -57,7 +48,7 @@ def _commonprefix(files):
 ################################################################################
 class CommitDialog(gnomeglade.Component):
     def __init__(self, parent):
-        gnomeglade.Component.__init__(self, paths.ui_dir("vcview.glade"), "commitdialog")
+        gnomeglade.Component.__init__(self, paths.ui_dir("vcview.ui"), "commitdialog")
         self.parent = parent
         self.widget.set_transient_for( parent.widget.get_toplevel() )
         selected = parent._get_selected_files()
@@ -88,7 +79,9 @@ COL_LOCATION, COL_STATUS, COL_REVISION, COL_TAG, COL_OPTIONS, COL_END = range(tr
 
 class VcTreeStore(tree.DiffTreeStore):
     def __init__(self):
-        tree.DiffTreeStore.__init__(self, 1, COL_END)
+        ntree = 1
+        types = [str] * COL_END * ntree
+        tree.DiffTreeStore.__init__(self, ntree, types)
         self.textstyle[tree.STATE_MISSING] = '<span foreground="#000088" strikethrough="true" weight="bold">%s</span>'
 
 ################################################################################
@@ -119,19 +112,19 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def __init__(self, prefs):
         melddoc.MeldDoc.__init__(self, prefs)
-        gnomeglade.Component.__init__(self, paths.ui_dir("vcview.glade"), "vcview")
+        gnomeglade.Component.__init__(self, paths.ui_dir("vcview.ui"), "vcview")
 
         actions = (
             ("VcCompare",       gtk.STOCK_DIALOG_INFO,      _("_Compare"),      None, _("Compare selected"), self.on_button_diff_clicked),
             ("VcOpen",          gtk.STOCK_OPEN,             None,               None, _("Open selected"), self.on_button_open_clicked),
-            ("VcCommit",        "vc-commit-24",             _("_Commit"),       None, _("Commit"), self.on_button_commit_clicked), # FIXME: popup used to use gtk.STOCK_GO_BACK
-            ("VcUpdate",        "vc-update-24",             _("_Update"),       None, _("Update"), self.on_button_update_clicked), # FIXME: popup used to use gtk.STOCK_GO_FORWARD
-            ("VcAdd",           "vc-add-24",                _("_Add"),          None, _("Add to VC"), self.on_button_add_clicked), # FIXME: popup used to use gtk.STOCK_ADD
-            ("VcAddBinary",     gtk.STOCK_ADD,              _("Add _Binary"),   None, _("Add binary to VC"), self.on_button_add_binary_clicked), # FIXME: stock is inconsistent with other VC actions
-            ("VcRemove",        "vc-remove-24",             _("_Remove"),       None, _("Remove from VC"), self.on_button_remove_clicked), # FIXME: popup used to use gtk.STOCK_REMOVE
+            ("VcCommit",        "vc-commit-24",             _("Co_mmit"),       None, _("Commit"), self.on_button_commit_clicked),
+            ("VcUpdate",        "vc-update-24",             _("_Update"),       None, _("Update"), self.on_button_update_clicked),
+            ("VcAdd",           "vc-add-24",                _("_Add"),          None, _("Add to VC"), self.on_button_add_clicked),
+            ("VcAddBinary",     None,                       _("Add _Binary"),   None, _("Add binary to VC"), self.on_button_add_binary_clicked),
+            ("VcRemove",        "vc-remove-24",             _("_Remove"),       None, _("Remove from VC"), self.on_button_remove_clicked),
             ("VcResolved",      "vc-resolve-24",            _("_Resolved"),     None, _("Mark as resolved for VC"), self.on_button_resolved_clicked),
             ("VcRevert",        gtk.STOCK_REVERT_TO_SAVED,  None,               None, _("Revert to original"), self.on_button_revert_clicked),
-            ("VcDeleteLocally", gtk.STOCK_DELETE,           None,               None, _("Delete locally"), self.on_button_delete_clicked), # FIXME: popup label was "_Remove locally"
+            ("VcDeleteLocally", gtk.STOCK_DELETE,           None,               None, _("Delete locally"), self.on_button_delete_clicked),
         )
 
         toggleactions = (
@@ -160,13 +153,17 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.treeview.set_model(self.model)
         self.treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
         self.treeview.set_headers_visible(1)
+        self.treeview.set_search_equal_func(self.treeview_search_cb)
         column = gtk.TreeViewColumn( _("Name") )
-        renpix = gtk.CellRendererPixbuf()
+        renicon = ui.emblemcellrenderer.EmblemCellRenderer()
         rentext = gtk.CellRendererText()
-        column.pack_start(renpix, expand=0)
+        column.pack_start(renicon, expand=0)
         column.pack_start(rentext, expand=1)
-        column.set_attributes(renpix, pixbuf=self.model.column_index(tree.COL_ICON, 0))
-        column.set_attributes(rentext, markup=self.model.column_index(tree.COL_TEXT, 0))
+        col_index = self.model.column_index
+        column.set_attributes(renicon,
+                              icon_name=col_index(tree.COL_ICON, 0),
+                              icon_tint=col_index(tree.COL_TINT, 0))
+        column.set_attributes(rentext, markup=col_index(tree.COL_TEXT, 0))
         self.treeview.append_column(column)
 
         def addCol(name, num):
@@ -213,6 +210,11 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.combobox_vcs.show()
         self.combobox_vcs.connect("changed", self.on_vc_change)
 
+    def on_container_switch_in_event(self, ui):
+        melddoc.MeldDoc.on_container_switch_in_event(self, ui)
+        # FIXME: Add real sensitivity handling
+        self.emit("next-diff-changed", True, True)
+
     def update_actions_sensitivity(self):
         """Disable actions that use not implemented VC plugin methods
         """
@@ -262,8 +264,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         if valid_vcs and default_active == -1:
             default_active = min(valid_vcs)
 
-        if gtk.pygtk_version >= (2, 12, 0):
-            self.combobox_vcs.set_tooltip_text(tooltip_texts[len(vcs) == 1])
+        self.combobox_vcs.set_tooltip_text(tooltip_texts[len(vcs) == 1])
         self.combobox_vcs.set_sensitive(len(vcs) > 1)
         self.combobox_vcs.lock = False
         self.combobox_vcs.set_active(default_active)
@@ -296,6 +297,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def recompute_label(self):
         self.label_text = os.path.basename(self.location)
+        # TRANSLATORS: This is the location of the directory the user is diffing
+        self.tooltip_text = _("%s: %s") % (_("Location"), self.location)
         self.label_changed()
 
     def _search_recursively_iter(self, iterstart):
@@ -345,7 +348,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 if len(entries) == 0:
                     self.model.add_empty(it, _("(Empty)"))
                 if differences or len(path)==1:
-                    _expand_to_root( self.treeview, path )
+                    self.treeview.expand_to_path(path)
             else: # just the root
                 self.treeview.expand_row( (0,), 0)
         self.vc.uncache_inventory()
@@ -375,30 +378,43 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             path = self.model.value_path(it, 0)
             self.run_diff( [path] )
 
-    def run_diff_iter(self, path_list, empty_patch_ok):
-        yield _("[%s] Fetching differences") % self.label_text
-        difffunc = self._command_iter(self.vc.diff_command(), path_list, 0).next
-        diff = None
-        while type(diff) != type(()):
-            diff = difffunc()
-            yield 1
-        prefix, patch = diff[0], diff[1]
-        yield _("[%s] Applying patch") % self.label_text
-        if patch:
-            self.show_patch(prefix, patch)
-        elif empty_patch_ok:
-            misc.run_dialog( _("No differences found."), parent=self, messagetype=gtk.MESSAGE_INFO)
-        else:
-            for path in path_list:
-                self.emit("create-diff", [path])
+    def run_diff_iter(self, path_list):
+        silent_error = hasattr(self.vc, 'switch_to_external_diff')
+        retry_diff = True
+        while retry_diff:
+            retry_diff = False
 
-    def run_diff(self, path_list, empty_patch_ok=0):
+            yield _("[%s] Fetching differences") % self.label_text
+            difffunc = self._command_iter(self.vc.diff_command(),
+                                          path_list, 0).next
+            diff = None
+            while type(diff) != type(()):
+                diff = difffunc()
+                yield 1
+            prefix, patch = diff[0], diff[1]
+            yield _("[%s] Applying patch") % self.label_text
+            if patch:
+                applied = self.show_patch(prefix, patch, silent=silent_error)
+                if not applied and silent_error:
+                    silent_error = False
+                    self.vc.switch_to_external_diff()
+                    retry_diff = True
+            else:
+                for path in path_list:
+                    self.emit("create-diff", [path])
+
+    def run_diff(self, path_list):
         for path in path_list:
-            self.scheduler.add_task(self.run_diff_iter([path], empty_patch_ok).next, atfront=1)
+            self.scheduler.add_task(self.run_diff_iter([path]).next, atfront=1)
+
+    def on_treeview_popup_menu(self, treeview):
+        time = gtk.get_current_event_time()
+        self.popup_menu.popup(None, None, None, 0, time)
+        return True
 
     def on_button_press_event(self, text, event):
-        if event.button==3:
-            self.popup_menu.popup(None, None, None, 3, event.time)
+        if event.button == 3:
+            self.popup_menu.popup(None, None, None, event.button, event.time)
             return len(self._get_selected_paths()) != 1
         return 0
 
@@ -498,12 +514,12 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_diff_clicked(self, obj):
         files = self._get_selected_files()
         if len(files):
-            self.run_diff(files, empty_patch_ok=1)
+            self.run_diff(files)
 
     def on_button_open_clicked(self, obj):
         self._open_files(self._get_selected_files())
 
-    def show_patch(self, prefix, patch):
+    def show_patch(self, prefix, patch, silent=False):
         tmpdir = tempfile.mkdtemp("-meld")
         self.tempdirs.append(tmpdir)
 
@@ -521,11 +537,18 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 open(destfile,"w").close()
             diffs.append( (destfile, pathtofile) )
 
-        patchcmd = self.vc.patch_command( tmpdir )
-        if misc.write_pipe(patchcmd, patch) == 0:
+        patchcmd = self.vc.patch_command(tmpdir)
+        try:
+            result = misc.write_pipe(patchcmd, patch, error=misc.NULL)
+        except OSError:
+            result = 1
+
+        if result == 0:
             for d in diffs:
+                os.chmod(d[0], 0444)
                 self.emit("create-diff", d)
-        else:
+            return True
+        elif not silent:
             import meldapp
             msg = _("""
                     Invoking 'patch' failed.
@@ -555,6 +578,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                             " ".join(patchcmd))
             msg = '\n'.join([line.strip() for line in msg.split('\n')])
             misc.run_dialog(msg, parent=self)
+        return False
 
     def refresh(self):
         self.set_location( self.model.value_path( self.model.get_iter_root(), 0 ) )
@@ -586,6 +610,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         it = self.find_iter_by_name(filename)
         if it:
             path = self.model.value_path(it, 0)
+            self.vc.update_file_state(path)
             files = self.vc.lookup_files([], [(os.path.basename(path), path)])[1]
             for e in files:
                 if e.path == path:
@@ -641,7 +666,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
         search = {gtk.gdk.SCROLL_UP : self.model.inorder_search_up}.get(direction, self.model.inorder_search_down)
         for it in search( start_iter ):
-            state = int(self.model.get_state( it, 0))
+            state = self.model.get_state(it, 0)
             if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
                 curpath = self.model.get_path(it)
                 self.treeview.expand_to_path(curpath)
@@ -650,3 +675,26 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_reload_activate(self, *extra):
         self.on_fileentry_activate(self.fileentry)
+
+    def on_find_activate(self, *extra):
+        self.treeview.emit("start-interactive-search")
+
+    def treeview_search_cb(self, model, column, key, it):
+        """Callback function for searching in VcView treeview"""
+        path = model.get_value(it, tree.COL_PATH)
+
+        # if query text contains slash, search in full path
+        if key.find('/') >= 0:
+            lineText = path
+        else:
+            lineText = os.path.basename(path)
+
+        # Perform case-insensitive matching if query text is all lower-case
+        if key.islower():
+            lineText = lineText.lower()
+
+        if lineText.find(key) >= 0:
+            # line matches
+            return False
+        else:
+            return True
